import abc
import uuid
import socket
import datetime

from coilmq.exception import ProtocolError, AuthError
from coilmq.util import frames
from coilmq.util.frames import Frame, ErrorFrame, ReceiptFrame, ConnectedFrame
from coilmq.util.concurrency import CoilThreadingTimer

SEND = 'SEND'
CONNECT = 'CONNECT'
MESSAGE = 'MESSAGE'
ERROR = 'ERROR'
CONNECTED = 'CONNECTED'
SUBSCRIBE = 'SUBSCRIBE'
UNSUBSCRIBE = 'UNSUBSCRIBE'
BEGIN = 'BEGIN'
COMMIT = 'COMMIT'
ABORT = 'ABORT'
ACK = 'ACK'
DISCONNECT = 'DISCONNECT'

VALID_COMMANDS = ['message', 'connect', 'connected', 'error', 'send',
                  'subscribe', 'unsubscribe', 'begin', 'commit', 'abort', 'ack', 'disconnect', 'nack', 'stomp']


class STOMP(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self, engine):
        self.engine = engine

    def stomp(self, frame):
        self.connect(frame)

    @abc.abstractmethod
    def process_frame(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def connect(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def send(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def subscribe(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def unsubscribe(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def begin(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def commit(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def abort(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def ack(self, frame):
        raise NotImplemented

    @abc.abstractmethod
    def disconnect(self, frame):
        raise NotImplemented


class STOMP10(STOMP):

    def process_frame(self, frame):
        """
        Dispatches a received frame to the appropriate internal method.

        @param frame: The frame that was received.
        @type frame: C{stompclient.frame.Frame}
        """
        cmd_method = frame.cmd.lower()

        if not cmd_method in VALID_COMMANDS:
            raise ProtocolError("Invalid STOMP command: {}".format(frame.cmd))

        method = getattr(self, cmd_method, None)

        if not self.engine.connected and method not in (self.connect, self.stomp):
            raise ProtocolError("Not connected.")

        try:
            transaction = frame.headers.get('transaction')
            if not transaction or method in (self.begin, self.commit, self.abort):
                method(frame)
            else:
                if not transaction in self.engine.transactions:
                    raise ProtocolError(
                        "Invalid transaction specified: %s" % transaction)
                self.engine.transactions[transaction].append(frame)
        except Exception as e:
            self.engine.log.error("Error processing STOMP frame: %s" % e)
            self.engine.log.exception(e)
            try:
                self.engine.connection.send_frame(ErrorFrame(str(e), str(e)))
            except Exception as e:
                self.engine.log.error("Could not send error frame: %s" % e)
                self.engine.log.exception(e)
        else:
            # The protocol is not especially clear here (not sure why I'm surprised)
            # about the expected behavior WRT receipts and errors.  We will assume that
            # the RECEIPT frame should not be sent if there is an error frame.
            # Also we'll assume that a transaction should not preclude sending the receipt
            # frame.
            # import pdb; pdb.set_trace()
            if frame.headers.get('receipt') and method != self.connect:
                self.engine.connection.send_frame(ReceiptFrame(
                    receipt=frame.headers.get('receipt')))

    def connect(self, frame, response=None):
        """
        Handle CONNECT command: Establishes a new connection and checks auth (if applicable).
        """
        self.engine.log.debug("CONNECT")

        if self.engine.authenticator:
            login = frame.headers.get('login')
            passcode = frame.headers.get('passcode')
            if not self.engine.authenticator.authenticate(login, passcode):
                raise AuthError("Authentication failed for %s" % login)

        self.engine.connected = True

        response = response or Frame(frames.CONNECTED)
        response.headers['session'] = uuid.uuid4()

        # TODO: Do we want to do anything special to track sessions?
        # (Actually, I don't think the spec actually does anything with this at all.)
        self.engine.connection.send_frame(response)

    def send(self, frame):
        """
        Handle the SEND command: Delivers a message to a queue or topic (default).
        """
        dest = frame.headers.get('destination')
        if not dest:
            raise ProtocolError('Missing destination for SEND command.')

        if dest.startswith('/queue/'):
            self.engine.queue_manager.send(frame)
        else:
            self.engine.topic_manager.send(frame)

    def subscribe(self, frame):
        """
        Handle the SUBSCRIBE command: Adds this connection to destination.
        """
        ack = frame.headers.get('ack')
        reliable = ack and ack.lower() == 'client'

        self.engine.connection.reliable_subscriber = reliable

        dest = frame.headers.get('destination')
        if not dest:
            raise ProtocolError('Missing destination for SUBSCRIBE command.')

        if dest.startswith('/queue/'):
            self.engine.queue_manager.subscribe(self.engine.connection, dest)
        else:
            self.engine.topic_manager.subscribe(self.engine.connection, dest)

    def unsubscribe(self, frame):
        """
        Handle the UNSUBSCRIBE command: Removes this connection from destination.
        """
        dest = frame.headers.get('destination')
        if not dest:
            raise ProtocolError('Missing destination for UNSUBSCRIBE command.')

        if dest.startswith('/queue/'):
            self.engine.queue_manager.unsubscribe(self.engine.connection, dest)
        else:
            self.engine.topic_manager.unsubscribe(self.engine.connection, dest)

    def begin(self, frame):
        """
        Handles BEGING command: Starts a new transaction.
        """
        if not frame.transaction:
            raise ProtocolError("Missing transaction for BEGIN command.")

        self.engine.transactions[frame.transaction] = []

    def commit(self, frame):
        """
        Handles COMMIT command: Commits specified transaction.
        """
        if not frame.transaction:
            raise ProtocolError("Missing transaction for COMMIT command.")

        if not frame.transaction in self.engine.transactions:
            raise ProtocolError("Invalid transaction: %s" % frame.transaction)

        for tframe in self.engine.transactions[frame.transaction]:
            del tframe.headers['transaction']
            self.process_frame(tframe)

        self.engine.queue_manager.clear_transaction_frames(
            self.engine.connection, frame.transaction)
        del self.engine.transactions[frame.transaction]

    def abort(self, frame):
        """
        Handles ABORT command: Rolls back specified transaction.
        """
        if not frame.transaction:
            raise ProtocolError("Missing transaction for ABORT command.")

        if not frame.transaction in self.engine.transactions:
            raise ProtocolError("Invalid transaction: %s" % frame.transaction)

        self.engine.queue_manager.resend_transaction_frames(
            self.engine.connection, frame.transaction)
        del self.engine.transactions[frame.transaction]

    def ack(self, frame):
        """
        Handles the ACK command: Acknowledges receipt of a message.
        """
        if not frame.message_id:
            raise ProtocolError("No message-id specified for ACK command.")
        self.engine.queue_manager.ack(self.engine.connection, frame)

    def disconnect(self, frame):
        """
        Handles the DISCONNECT command: Unbinds the connection.

        Clients are supposed to send this command, but in practice it should not be
        relied upon.
        """
        self.engine.log.debug("Disconnect")
        self.engine.unbind()


class STOMP11(STOMP10):

    SUPPORTED_VERSIONS = {'1.0', '1.1'}

    def __init__(self, engine, send_heartbeat_interval=100, receive_heartbeat_interval=100, *args, **kwargs):
        super(STOMP11, self).__init__(engine)
        self.last_hb = datetime.datetime.now()
        self.last_hb_sent = datetime.datetime.now()
        self.timer = CoilThreadingTimer()

        # flags to control heartbeating
        self.send_hb = self.receive_hb = False

        self.send_heartbeat_interval = datetime.timedelta(milliseconds=send_heartbeat_interval)
        self.receive_heartbeat_interval = datetime.timedelta(milliseconds=receive_heartbeat_interval)

    def enable_heartbeat(self, cx, cy, response):
        if self.send_heartbeat_interval and cy:
            self.send_heartbeat_interval = max(self.send_heartbeat_interval, datetime.timedelta(milliseconds=cy))
            self.timer.schedule(max(self.send_heartbeat_interval, datetime.timedelta(milliseconds=cy)).total_seconds(), self.send_heartbeat)
        if self.receive_heartbeat_interval and cx:
            self.timer.schedule(max(self.send_heartbeat_interval, datetime.timedelta(milliseconds=cx)).total_seconds(),
                                self.receive_heartbeat)
        self.timer.start()
        response.headers['heart-beat'] = '{0},{1}'.format(int(self.send_heartbeat_interval.microseconds / 1000),
                                                          int(self.receive_heartbeat_interval.microseconds / 1000))

    def disable_heartbeat(self):
        self.timer.stop()

    def send_heartbeat(self):
        # screw it, just send an error frame
        self.engine.connection.send_frame(ErrorFrame('heartbeat'))

    def receive_heartbeat(self):
        ago = datetime.datetime.now() - self.last_hb
        if ago > self.receive_heartbeat_interval:
            self.engine.log.debug("No heartbeat was received for {0} seconds".format(ago.total_seconds()))
            self.engine.unbind()

    def connect(self, frame, response=None):
        connected_frame = Frame(frames.CONNECTED)
        self._negotiate_protocol(frame, connected_frame)
        heart_beat = frame.headers.get('heart-beat', '0,0')
        if heart_beat:
            self.enable_heartbeat(*map(int, heart_beat.split(',')), response=connected_frame)
        super(STOMP11, self).connect(frame, response=connected_frame)

    def nack(self, frame):
        """
        Handles the NACK command: Unacknowledges receipt of a message.
        For now, this is just a placeholder to implement this version of the protocol
        """
        if not frame.headers.get('message-id'):
            raise ProtocolError("No message-id specified for NACK command.")
        if not frame.headers.get('subscription'):
            raise ProtocolError("No subscription specified for NACK command.")

    def _negotiate_protocol(self, frame, response):
        client_versions = frame.headers.get('accept-version')
        if not client_versions:
            raise ProtocolError('No version specified')
        common = set(client_versions.split(',')) & self.SUPPORTED_VERSIONS
        if not common:
            versions = ','.join(self.SUPPORTED_VERSIONS)
            self.engine.connection.send_frame(Frame(
                    frames.ERROR,
                    headers={'version': versions, 'content-type': frames.TEXT_PLAIN},
                    body='Supported protocol versions are {0}'.format(versions)
            ))
        else:
            response.headers['version'] = max(common)
            protocol_class = PROTOCOL_MAP[response.headers['version']]
            if type(self) is not protocol_class:
                self.engine.protocol = protocol_class(self.engine)
                self.engine.protocol.connect(frame, response=response)


class STOMP12(STOMP11):

    SUPPORTED_VERSIONS = STOMP11.SUPPORTED_VERSIONS.union({'1.2', })

    def connect(self, frame, response=None):
        host = frame.headers.get('host')
        if not host:
            raise ProtocolError('"host" header is required')
        if host != socket.getfqdn():
            raise ProtocolError('Virtual hosting is not supported or host is unknown')
        super(STOMP12, self).connect(frame, response)


PROTOCOL_MAP = {'1.0': STOMP10, '1.1': STOMP11, '1.2': STOMP12}



#!python
"""
Entrypoint for starting the application.
"""
import logging
import time
import threading
from contextlib import contextmanager

import daemon as pydaemon
import pid
import click

from coilmq.config import config as global_config, init_config, init_logging, resolve_name
from coilmq.protocol import STOMP11
from coilmq.topic import TopicManager
from coilmq.queue import QueueManager
from coilmq.server.socket_server import ThreadedStompServer

__authors__ = ['"Hans Lellelid" <hans@xmpl.org>']
__copyright__ = "Copyright 2009 Hans Lellelid"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
 
  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

logger = logging.getLogger(__name__)


def server_from_config(config=None, server_class=None, additional_kwargs=None):
    """
    Gets a configured L{coilmq.server.StompServer} from specified config.

    If `config` is None, global L{coilmq.config.config} var will be used instead.

    The `server_class` and `additional_kwargs` are primarily hooks for using this method
    from a testing environment.

    @param config: A C{ConfigParser.ConfigParser} instance with loaded config values.
    @type config: C{ConfigParser.ConfigParser}

    @param server_class: Which class to use for the server.  (This doesn't come from config currently.)
    @type server_class: C{class}

    @param additional_kwargs: Any additional args that should be passed to class.
    @type additional_kwargs: C{list}

    @return: The configured StompServer.
    @rtype: L{coilmq.server.StompServer}
    """
    global global_config
    if not config:
        config = global_config

    queue_store_factory = resolve_name(config.get('coilmq', 'qstore.factory'))
    subscriber_scheduler_factory = resolve_name(config.get(
        'coilmq', 'scheduler.subscriber_priority_factory'))
    queue_scheduler_factory = resolve_name(config.get(
        'coilmq', 'scheduler.queue_priority_factory'))

    if config.has_option('coilmq', 'auth.factory'):
        authenticator_factory = resolve_name(
            config.get('coilmq', 'auth.factory'))
        authenticator = authenticator_factory()
    else:
        authenticator = None

    server = ThreadedStompServer((config.get('coilmq', 'listen_addr'), config.getint('coilmq', 'listen_port')),
                                 queue_manager=QueueManager(store=queue_store_factory(),
                                                            subscriber_scheduler=subscriber_scheduler_factory(),
                                                            queue_scheduler=queue_scheduler_factory()),
                                 topic_manager=TopicManager(),
                                 authenticator=authenticator,
                                 protocol=STOMP11)
    logger.info("Created server:%r" % server)
    return server


def context_serve(context, configfile, listen_addr, listen_port, logfile,
                  debug, daemon, uid, gid, pidfile, umask, rundir):
    """
    Takes a context object, which implements the __enter__/__exit__ "with" interface 
    and starts a server within that context.

    This method is a refactored single-place for handling the server-run code whether
    running in daemon or non-daemon mode.  It is invoked with a dummy (passthrough) 
    context object for the non-daemon use case. 

    @param options: The compiled collection of options that need to be parsed. 
    @type options: C{ConfigParser}

    @param context: The context object that implements __enter__/__exit__ "with" methods.
    @type context: C{object}

    @raise Exception: Any underlying exception will be logged but then re-raised.
    @see: server_from_config()
    """
    global global_config

    server = None
    try:
        with context:
            # There's a possibility here that init_logging() will throw an exception.  If it does,
            # AND we're in a daemon context, then we're not going to be able to do anything with it.
            # We've got no stderr/stdout here; and so (to my knowledge) no reliable (& cross-platform),
            # way to display errors.
            level = logging.DEBUG if debug else logging.INFO
            init_logging(logfile=logfile, loglevel=level,
                         configfile=configfile)

            server = server_from_config()
            logger.info("Stomp server listening on %s:%s" % server.server_address)

            if debug:
                poll_interval = float(global_config.get(
                    'coilmq', 'debug.stats_poll_interval'))
                if poll_interval:  # Setting poll_interval to 0 effectively disables it.
                    def diagnostic_loop(server):
                        log = logger
                        while True:
                            log.debug(
                                "Stats heartbeat -------------------------------")
                            store = server.queue_manager.store
                            for dest in store.destinations():
                                log.debug("Queue %s: size=%s, subscribers=%s" % (
                                    dest, store.size(dest), server.queue_manager.subscriber_count(dest)))

                            # TODO: Add number of subscribers?

                            time.sleep(poll_interval)

                    diagnostic_thread = threading.Thread(
                        target=diagnostic_loop, name='DiagnosticThread', args=(server,))
                    diagnostic_thread.daemon = True
                    diagnostic_thread.start()

            server.serve_forever()

    except (KeyboardInterrupt, SystemExit):
        logger.info("Stomp server stopped by user interrupt.")
        raise SystemExit()
    except Exception as e:
        logger.error("Stomp server stopped due to error: %s" % e)
        logger.exception(e)
        raise SystemExit()
    finally:
        if server:
            server.server_close()


def _main(config=None, host=None, port=None, logfile=None, debug=None,
          daemon=None, uid=None, gid=None, pidfile=None, umask=None, rundir=None):

    # Note that we must initialize the configuration before we enter the context
    # block; however, we _cannot_ initialize logging until we are in the context block
    # (so we defer that until the context_serve call.)
    init_config(config)

    if host is not None:
        global_config.set('coilmq', 'listen_addr', host)

    if port is not None:
        global_config.set('coilmq', 'listen_port', str(port))

    # in an on-daemon mode, we use a dummy context objectx
    # so we can use the same run-server code as the daemon version.
    context = pydaemon.DaemonContext(uid=uid,
                                     gid=gid,
                                     pidfile=pid.PidFile(pidname=pidfile) if pidfile else None,
                                     umask=int(umask, 8),
                                     working_directory=rundir) if daemon else contextmanager(lambda: (yield))()

    context_serve(context, config, host, port, logfile, debug, daemon, uid, gid, pidfile, umask, rundir)


@click.command()
@click.option("-c", "--config", help="Read configuration from FILE. (CLI options override config file.)", metavar="FILE")
@click.option("-b", "--host", help="Listen on specified address (default 127.0.0.1)", metavar="ADDR")
@click.option("-p", "--port", help="Listen on specified port (default 61613)", type=int, metavar="PORT")
@click.option("-l", "--logfile", help="Log to specified file (unless logging configured in config file).", metavar="FILE")
@click.option("--debug", default=False, help="Sets logging to debug (unless logging configured in config file).")
@click.option("-d", "--daemon", default=False, help="Run server as a daemon (default False).")
@click.option("-u", "--uid", help="The user/UID to use for daemon process.", metavar="UID")
@click.option("-g", "--gid", help="The group/GID to use for daemon process.", metavar="GID")
@click.option("--pidfile",   help="The PID file to use.", metavar="FILE")
@click.option("--umask", help="Umask (octal) to apply for daemonized process.", metavar="MASK")
@click.option('--rundir', help="The working directory to use for the daemonized process (default /).", metavar="DIR")
def main(config, host, port, logfile, debug, daemon, uid, gid, pidfile, umask, rundir):
    """
    Main entry point for running a socket server from the commandline.

    This method will read in options from the commandline and call the L{config.init_config} method
    to get everything setup.  Then, depending on whether deamon mode was specified or not, 
    the process may be forked (or not) and the server will be started.
    """

    _main(**locals())


if __name__ == '__main__':
    try:
        main()
    except (KeyboardInterrupt, SystemExit):
        pass
    except Exception as e:
        logger.error("Server terminated due to error: %s" % e)
        logger.exception(e)



from __future__ import absolute_import

from coilmq.protocol import STOMP10

"""
Core STOMP server logic, abstracted from socket transport implementation.

While this is abstracted from the socket transport implementation, it does
operate on the assumption that there is an available connection to send response
frames.

We're also making some simplified assumptions here which may make this engine
impractical for [high-performance] use in specifically asynchronous frameworks.
More specifically, this class was not explicitly designed around async patterns, 
meaning that it would likely be problematic to use with a framework like Twisted 
if the underlying storage implementations were processor intensive (e.g. database
access).  For the default memory storage engines, this shouldn't be a problem.

This code is inspired by the design of the Ruby stompserver project, by 
Patrick Hurley and Lionel Bouton.  See http://stompserver.rubyforge.org/
"""
import logging
from collections import defaultdict

__authors__ = ['"Hans Lellelid" <hans@xmpl.org>']
__copyright__ = "Copyright 2009 Hans Lellelid"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
 
  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


class StompEngine(object):
    """ 
    The engine provides the core business logic that we use to respond to STOMP protocol
    messages.  

    This class is transport-agnostic; it exposes methods that expect STOMP frames and
    uses the attached connection to send frames to connected clients.

    @ivar connection: The connection (aka "protocol") backing this engine.
    @type connection: L{coilmq.server.StompConnection}

    @ivar authenticator: An C{Authenticator} implementation to use.  Setting this value
                            will implicitly cause authentication to be required.
    @type authenticator: L{coilmq.auth.Authenticator}

    @ivar queue_manager: The C{QueueManager} implementation to use.
    @type queue_manager: L{coilmq.queue.QueueManager}

    @ivar topic_manager: The C{TopicManager} implementation to use.
    @type topic_manager: L{coilmq.topic.TopicManager}

    @ivar transactions: Active transactions for this connection.
    @type transactions: C{dict} of C{str} to C{list} 

    @ivar connected: Whether engine is connected.
    @type connected: C{bool}
    """

    def __init__(self, connection, authenticator, queue_manager, topic_manager, protocol=STOMP10):
        """
        @param connection: The stomp connection backing this engine.
        @type connection: L{coilmq.server.StompConnection}
        """
        self.log = logging.getLogger('%s.%s' % (
            self.__class__.__module__, self.__class__.__name__))
        self.connection = connection
        self.authenticator = authenticator
        self.queue_manager = queue_manager
        self.topic_manager = topic_manager
        self.connected = False
        self.transactions = defaultdict(list)

        self.protocol = protocol(self)

    def process_frame(self, frame):
        self.protocol.process_frame(frame)

    def unbind(self):
        """
        Unbinds this connection from queue and topic managers (freeing up resources)
        and resets state.
        """
        self.connected = False
        self.queue_manager.disconnect(self.connection)
        self.topic_manager.disconnect(self.connection)



"""
Exception classes used by CoilMQ.

CoilMQ exceptions extend C{RuntimeError} or other appropriate sub-classes.  These will be
thrown if there is not a more appropriate error class already provided by builtins.
"""
__authors__ = ['"Hans Lellelid" <hans@xmpl.org>']
__copyright__ = "Copyright 2009 Hans Lellelid"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
 
  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


class ProtocolError(RuntimeError):
    """
    Represents an error at the STOMP protocol layer.
    """


class ConfigError(RuntimeError):
    """
    Represents an error in the configuration of the application.
    """


class AuthError(RuntimeError):
    """
    Represents an authentication or authorization error.
    """


class ClientDisconnected(Exception):
    """
    A signal that client has disconnected (so we shouldn't try to keep reading from the client).
    """



try:
    import redis
except ImportError:
    import sys; sys.exit('please, install redis-py package to use redis-store')
import threading
try:
    import cPickle as pickle
except ImportError:
    import pickle

from coilmq.store import QueueStore
from coilmq.util.concurrency import synchronized
from coilmq.config import config

__authors__ = ('"Hans Lellelid" <hans@xmpl.org>', '"Alexander Zhukov" <zhukovaa90@gmail.com>')
__copyright__ = "Copyright 2009 Hans Lellelid"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""

lock = threading.RLock()


def make_redis_store(cfg=None):
    return RedisQueueStore(
        redis_conn=redis.Redis(**dict((cfg or config).items('redis'))))


class RedisQueueStore(QueueStore):
    """Simple Queue with Redis Backend"""
    def __init__(self, redis_conn=None):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db = redis_conn or redis.Redis()
        # self.key = '{0}:{1}'.format(namespace, name)
        super(RedisQueueStore, self).__init__()

    @synchronized(lock)
    def enqueue(self, destination, frame):
        self.__db.rpush(destination, pickle.dumps(frame))

    @synchronized(lock)
    def dequeue(self, destination):
        item = self.__db.lpop(destination)
        if item:
            return pickle.loads(item)

    @synchronized(lock)
    def requeue(self, destination, frame):
        self.enqueue(destination, frame)

    @synchronized(lock)
    def size(self, destination):
        return self.__db.llen(destination)

    @synchronized(lock)
    def has_frames(self, destination):
        return self.size(destination) > 0

    @synchronized(lock)
    def destinations(self):
        return self.__db.keys()



"""
Tools to facilitate developing thread-safe components.
"""

import abc
import threading
import functools

__authors__ = ['"Hans Lellelid" <hans@xmpl.org>']
__copyright__ = "Copyright 2009 Hans Lellelid"
__license__ = """Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""


def synchronized(lock):
    def synchronize(func):
        """
        Decorator to lock and unlock a method (Phillip J. Eby).

        This function is to be used with object instance methods; the object must
        have a _lock variable (of type C{threading.Lock} or C{threading.RLock}).

        @param func: Method to decorate
        @type func: C{callable}
        """
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with lock:
                return func(*args, **kwargs)
        return wrapper
    return synchronize


class CoilTimerBase(object):

    __metaclass__ = abc.ABCMeta

    def __init__(self):
        self.jobs = []

    def schedule(self, period, callback):
        self.jobs.append((period, callback))

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError

    @abc.abstractmethod
    def start(self):
        raise NotImplementedError

    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


# TODO: check against following notes
# <http://stackoverflow.com/questions/2124540/how-does-timer-in-python-work-regarding-mutlithreading>
# <http://stackoverflow.com/questions/12435211/python-threading-timer-repeat-function-every-n-seconds>
class CoilThreadingTimer(CoilTimerBase):

    def __init__(self, *args, **kwargs):
        super(CoilThreadingTimer, self).__init__(*args, **kwargs)
        self._running = False

    def run(self):
        def run_job(interval, callback):
            if self._running:
                threading.Timer(interval, run_job, args=(interval, callback)).start()
                callback()
        for period, job in self.jobs:
            run_job(period, job)

    def start(self):
        self._running = True
        self.run()

    def stop(self):
        self._running = False





"""
Simple, lightweight, and easily extensible STOMP message broker.
"""
import os.path

from setuptools import setup, find_packages
from setuptools.command.test import test


class PyTest(test):
    def finalize_options(self):
        test.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        pytest.main(self.test_args)


version = '1.0.0'

long_description = """
The provided server implementation for CoilMQ uses the Python SocketServer
libraries; however, CoilMQ is only loosely coupled to this server
implementation.  It could be used with other socket implementations.

The CoilMQ core classes and bundled storage implementations are built to be
thread-safe.


"""


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


setup(
    name='CoilMQ',
    version=version,
    description=__doc__,
    long_description=long_description + read('docs/news.txt'),
    keywords='stomp server broker messaging',
    license='Apache',
    author='Hans Lellelid',
    author_email='hans@xmpl.org',
    url='https://github.com/hozn/coilmq',
    packages=find_packages(exclude=['ez_setup', 'distribute_setup', 'tests', 'tests.*']),
    package_dir={'coilmq':  'coilmq'},
    package_data={'coilmq': ['config/*.cfg*', 'tests/resources/*']},
    zip_safe=False,  # We use resource_filename for logging configuration and some unit tests.
    include_package_data=True,
    tests_require=read('requirements/test.txt').splitlines(),
    cmdclass={'test': PyTest},
    install_requires=read('requirements/prod.txt').splitlines(),
    extras_require={
        'sqlalchemy': ['SQLAlchemy'],
        'redis': ['redis']
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Communications",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Networking"
    ],
    entry_points="""
    [console_scripts]
    coilmq = coilmq.start:main
    """,
)



from functools import partial
import re
import logging
from collections import OrderedDict
import io

import six

SEND = 'SEND'
CONNECT = 'CONNECT'
MESSAGE = 'MESSAGE'
ERROR = 'ERROR'
CONNECTED = 'CONNECTED'
SUBSCRIBE = 'SUBSCRIBE'
UNSUBSCRIBE = 'UNSUBSCRIBE'
BEGIN = 'BEGIN'
COMMIT = 'COMMIT'
ABORT = 'ABORT'
ACK = 'ACK'
NACK = 'NACK'
DISCONNECT = 'DISCONNECT'

VALID_COMMANDS = ['message', 'connect', 'connected', 'error', 'send',
                  'subscribe', 'unsubscribe', 'begin', 'commit', 'abort', 'ack', 'disconnect', 'nack']

TEXT_PLAIN = 'text/plain'


class IncompleteFrame(Exception):
    """The frame has incomplete body"""


class BodyNotTerminated(Exception):
    """The frame's body is not terminated with the NULL character"""


class EmptyBuffer(Exception):
    """The buffer is empty"""


def parse_headers(buff):
    """
    Parses buffer and returns command and headers as strings
    """
    preamble_lines = list(map(
        lambda x: six.u(x).decode(),
        iter(lambda: buff.readline().strip(), b''))
    )
    if not preamble_lines:
        raise EmptyBuffer()
    return preamble_lines[0], OrderedDict([l.split(':') for l in preamble_lines[1:]])


def parse_body(buff, headers):
    content_length = int(headers.get('content-length', -1))
    body = buff.read(content_length)
    if content_length >= 0:
        if len(body) < content_length:
            raise IncompleteFrame()
        terminator = six.u(buff.read(1)).decode()
        if not terminator:
            raise BodyNotTerminated()
    else:
        # no content length
        body, terminator, rest = body.partition(b'\x00')
        if not terminator:
            raise BodyNotTerminated()
        else:
            buff.seek(-len(rest), 2)

    return body


class Frame(object):
    """
    A STOMP frame (or message).

    :param cmd: the protocol command
    :param headers: a map of headers for the frame
    :param body: the content of the frame.
    """

    def __init__(self, cmd, headers=None, body=None):
        self.cmd = cmd
        self.headers = headers or {}
        self.body = body or ''

    def __str__(self):
        return '{{cmd={0},headers=[{1}],body={2}}}'.format(
            self.cmd,
            self.headers,
            self.body if isinstance(
                self.body, six.binary_type) else six.b(self.body)
        )

    def __eq__(self, other):
        """ Override equality checking to test for matching command, headers, and body. """
        return all([isinstance(other, Frame),
                    self.cmd == other.cmd,
                    self.headers == other.headers,
                    self.body == other.body])

    @property
    def transaction(self):
        return self.headers.get('transaction')

    @classmethod
    def from_buffer(cls, buff):
        cmd, headers = parse_headers(buff)
        body = parse_body(buff, headers)
        return cls(cmd, headers=headers, body=body)

    def pack(self):
        """
        Create a string representation from object state.

        @return: The string (bytes) for this stomp frame.
        @rtype: C{str}
        """

        self.headers.setdefault('content-length', len(self.body))

        # Convert and append any existing headers to a string as the
        # protocol describes.
        headerparts = ("{0}:{1}\n".format(key, value)
                       for key, value in self.headers.items())

        # Frame is Command + Header + EOF marker.
        return six.b("{0}\n{1}\n".format(self.cmd, "".join(headerparts))) + (self.body if isinstance(self.body, six.binary_type) else six.b(self.body)) + six.b('\x00')


class ConnectedFrame(Frame):
    """ A CONNECTED server frame (response to CONNECT).

    @ivar session: The (throw-away) session ID to include in response.
    @type session: C{str}
    """

    def __init__(self, session, extra_headers=None):
        """
        @param session: The (throw-away) session ID to include in response.
        @type session: C{str}
        """
        super(ConnectedFrame, self).__init__(
            cmd='connected', headers=extra_headers or {})
        self.headers['session'] = session


class HeaderValue(object):
    """
    An descriptor class that can be used when a calculated header value is needed.

    This class is a descriptor, implementing  __get__ to return the calculated value.
    While according to  U{http://docs.codehaus.org/display/STOMP/Character+Encoding} there
    seems to some general idea about having UTF-8 as the character encoding for headers;
    however the C{stomper} lib does not support this currently.

    For example, to use this class to generate the content-length header:

        >>> body = 'asdf'
        >>> headers = {}
        >>> headers['content-length'] = HeaderValue(calculator=lambda: len(body))
        >>> str(headers['content-length'])
        '4'

    @ivar calc: The calculator function.
    @type calc: C{callable}
    """

    def __init__(self, calculator):
        """
        @param calculator: The calculator callable that will yield the desired value.
        @type calculator: C{callable}
        """
        if not callable(calculator):
            raise ValueError("Non-callable param: %s" % calculator)
        self.calc = calculator

    def __get__(self, obj, objtype):
        return self.calc()

    def __str__(self):
        return str(self.calc())

    def __set__(self, obj, value):
        self.calc = value

    def __repr__(self):
        return '<%s calculator=%s>' % (self.__class__.__name__, self.calc)


class ErrorFrame(Frame):
    """ An ERROR server frame. """

    def __init__(self, message, body=None, extra_headers=None):
        """
        @param body: The message body bytes.
        @type body: C{str}
        """
        super(ErrorFrame, self).__init__(cmd='error',
                                         headers=extra_headers or {}, body=body)
        self.headers['message'] = message
        self.headers[
            'content-length'] = HeaderValue(calculator=lambda: len(self.body))

    def __repr__(self):
        return '<%s message=%r>' % (self.__class__.__name__, self.headers['message'])


class ReceiptFrame(Frame):
    """ A RECEIPT server frame. """

    def __init__(self, receipt, extra_headers=None):
        """
        @param receipt: The receipt message ID.
        @type receipt: C{str}
        """
        super(ReceiptFrame, self).__init__(
            'RECEIPT', headers=extra_headers or {})
        self.headers['receipt-id'] = receipt


class FrameBuffer(object):
    """
    A customized version of the StompBuffer class from Stomper project that returns frame objects
    and supports iteration.

    This version of the parser also assumes that stomp messages with no content-lengh
    end in a simple \\x00 char, not \\x00\\n as is assumed by
    C{stomper.stompbuffer.StompBuffer}. Additionally, this class differs from Stomper version
    by conforming to PEP-8 coding style.

    This class can be used to smooth over a transport that may provide partial frames (or
    may provide multiple frames in one data buffer).

    @ivar _buffer: The internal byte buffer.
    @type _buffer: C{str}

    @ivar debug: Log extra parsing debug (logs will be DEBUG level).
    @type debug: C{bool}
    """

    # regexp to check that the buffer starts with a command.
    command_re = re.compile('^(.+?)\n')

    # regexp to remove everything up to and including the first
    # instance of '\x00' (used in resynching the buffer).
    sync_re = re.compile('^.*?\x00')

    # regexp to determine the content length. The buffer should always start
    # with a command followed by the headers, so the content-length header will
    # always be preceded by a newline.  It may not always proceeded by a
    # newline, though!
    content_length_re = re.compile('\ncontent-length\s*:\s*(\d+)\s*(\n|$)')

    def __init__(self):
        self._buffer = io.BytesIO()
        self._pointer = 0
        self.debug = False
        self.log = logging.getLogger('%s.%s' % (
            self.__module__, self.__class__.__name__))

    def clear(self):
        """
        Clears (empties) the internal buffer.
        """
        self._buffer = io

    def buffer_len(self):
        """
        @return: Number of bytes in the internal buffer.
        @rtype: C{int}
        """
        return len(self._buffer)

    def buffer_empty(self):
        """
        @return: C{True} if buffer is empty, C{False} otherwise.
        @rtype: C{bool}
        """
        return not bool(self._buffer)

    def append(self, data):
        """
        Appends bytes to the internal buffer (may or may not contain full stomp frames).

        @param data: The bytes to append.
        @type data: C{str}
        """
        self._buffer.write(data)

    def extract_frame(self):
        """
        Pulls one complete frame off the buffer and returns it.

        If there is no complete message in the buffer, returns None.

        Note that the buffer can contain more than once message. You
        should therefore call this method in a loop (or use iterator
        functionality exposed by class) until None returned.

        @return: The next complete frame in the buffer.
        @rtype: L{stomp.frame.Frame}
        """
        # (mbytes, hbytes) = self._find_message_bytes(self.buffer)
        # if not mbytes:
        #     return None
        #
        # msgdata = self.buffer[:mbytes]
        # self.buffer = self.buffer[mbytes:]
        # hdata = msgdata[:hbytes]
        # # Strip off any leading whitespace from headers; this is necessary, because
        # # we do not (any longer) expect a trailing \n after the \x00 byte (which means
        # # it will become a leading \n to the next frame).
        # hdata = hdata.lstrip()
        # elems = hdata.split('\n')
        # cmd = elems.pop(0)
        # headers = {}
        #
        # for e in elems:
        #     try:
        #         (k,v) = e.split(':', 1) # header values may contain ':' so specify maxsplit
        #     except ValueError:
        #         continue
        #     headers[k.strip()] = v.strip()
        #
        # # hbytes points to the start of the '\n\n' at the end of the header,
        # # so 2 bytes beyond this is the start of the body. The body EXCLUDES
        # # the final byte, which is  '\x00'.
        # body = msgdata[hbytes + 2:-1]
        self._buffer.seek(self._pointer, 0)
        try:
            f = Frame.from_buffer(self._buffer)
            self._pointer = self._buffer.tell()
        except (IncompleteFrame, EmptyBuffer):
            self._buffer.seek(self._pointer, 0)
            return None

        return f

    def __iter__(self):
        """
        Returns an iterator object.
        """
        return self

    def __next__(self):
        """
        Return the next STOMP message in the buffer (supporting iteration).

        @rtype: L{stomp.frame.Frame}
        """
        msg = self.extract_frame()
        if not msg:
            raise StopIteration()
        return msg

    def next(self):
        return self.__next__()
